



// v4.0
//##############################################################################
//#################                HEAP                       ##################
//##############################################################################


        (*
          DEFINE UPDATE_HEAPSIZE TO KEEP THE HEAPSIZE UPDATED.
        *)
        {$DEFINE UPDATE_HEAPSIZE}


//##############################################################################
//#################                HEAP                       ##################
//##############################################################################
Unit AHeap;




        		//###############
        		//## INTERFACE ##
        		//###############
        		Interface

        		//##############
        		//##   USES   ##
        		//##############
        		Uses
                        //######################################################
        		//######################## SYSTEM ######################
                                		  {System,}
                                		  HWindows;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                                INTERFACE                                 ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




// THE GRANULARITY OF HEAP ALLOCATIONS IN WIN32 IS 16 BYTES. SO IF YOU
// REQUEST A GLOBAL MEMORY ALLOCATION OF 1 BYTE, THE HEAP RETURNS A POINTER
// TO A CHUNK OF MEMORY, GUARANTEEING THAT THE 1 BYTE IS AVAILABLE. CHANCES
// ARE, 16 BYTES WILL ACTUALLY BE AVAILABLE BECAUSE THE HEAP CANNOT ALLOCATE
// LESS THAN 16 BYTES AT A TIME.




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                            EXPLICIT INTERFACE                            ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################




//##############################################################################
//##                              init and final                              ##
//##############################################################################

procedure Init;
procedure Final;

//##############################################################################
//##                         memory manager functions                         ##
//##############################################################################

function PointerSize(APointer: Pointer): Integer;
function HeapGet(ASize: Integer): Pointer;
function HeapFree(APointer: Pointer): Integer;
function HeapRealloc(APointer: Pointer; ASize: Integer): Pointer;

//##############################################################################
//##                     memory manipulation functions                        ##
//##############################################################################

// copies memory choosing the adequate mode to avoid overlapping.
procedure HeapMove(ADest, ASrc: Pointer; ADWordCount: Cardinal);
procedure HeapMoveByte(ADest, ASrc: Pointer; AByteCount: Cardinal);
//#
// copies memory from beginning to end.
procedure HeapCopy(ADest, ASrc: Pointer; ADWordCount: Cardinal);
procedure HeapCopyByte(ADest, ASrc: Pointer; AByteCount: Cardinal);
//#
// copies memory from end to beginning.
procedure HeapCopyRev(ADest, ASrc: Pointer; ADWordCount: Cardinal);
procedure HeapCopyByteRev(ADest, ASrc: Pointer; AByteCount: Cardinal);
//#
procedure HeapFill(ADest: Pointer; ADWordPattern: Cardinal; ADWordCount: Cardinal);
procedure HeapFillByte(ADest: Pointer; ABytePattern: Byte; AByteCount: Cardinal);
//#
function HeapCompare(ASrc, ADest: Pointer; ADWordCount: Cardinal): Boolean;
function HeapCompareByte(ASrc, ADest: Pointer; AByteCount: Cardinal): Boolean;




                        //##############
                        //###  VAR   ###
                        //##############
                        Var
                        //######################################################
                        //##################### GLOBAL USED ####################

  HeapHandle: Cardinal = 0;
  HeapSize: Integer = 0;

  HeapMemoryManager: TMemoryManager = (
                                       GetMem: HeapGet;
                                       FreeMem: HeapFree;
                                       ReallocMem: HeapRealloc
                                      );

  OldMemoryManager: TMemoryManager;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##########################                          ##########################
//##########################                          ##########################
//##                              IMPLEMENTATION                              ##
//##########################                          ##########################
//##########################                          ##########################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
Implementation




//##############################################################################
//##                         memory manager functions                         ##
//##############################################################################

function PointerSize(APointer: Pointer): Integer;
begin
        Result:= HWindows.HeapSize(HeapHandle, 0, APointer);
end;


function HeapGet(ASize: Integer): Pointer;
begin
        Result:= HWindows.HeapAlloc(HeapHandle, HEAP_ZERO_MEMORY, ASize);

        {$IFDEF UPDATE_HEAPSIZE}
        Inc(HeapSize, ASize);
        {$ENDIF}
end;


function HeapFree(APointer: Pointer): Integer;
begin
        {$IFDEF UPDATE_HEAPSIZE}
        Dec(HeapSize, PointerSize(APointer));
        {$ENDIF}

        Result:= Integer(not HWindows.HeapFree(HeapHandle, 0, APointer));
end;


function HeapRealloc(APointer: Pointer; ASize: Integer): Pointer;
begin
        {$IFDEF UPDATE_HEAPSIZE}
        HeapSize:= HeapSize + ASize - PointerSize(APointer);
        {$ENDIF}

        Result:= HWindows.HeapRealloc(HeapHandle, HEAP_ZERO_MEMORY, APointer, ASize);
end;

//##############################################################################
//##                     memory manipulation functions                        ##
//##############################################################################

procedure HeapMove(ADest, ASrc: Pointer; ADWordCount: Cardinal);
begin
        if (Integer(ADest) > Integer(ASrc)) then HeapCopy(ADest, ASrc, ADWordCount)
        else if (Integer(ADest) < Integer(ASrc)) then HeapCopyRev(ADest, ASrc, ADWordCount)
end;


procedure HeapMoveByte(ADest, ASrc: Pointer; AByteCount: Cardinal);
begin
        if (Integer(ADest) > Integer(ASrc)) then HeapCopyByte(ADest, ASrc, AByteCount)
        else if (Integer(ADest) < Integer(ASrc)) then HeapCopyByteRev(ADest, ASrc, AByteCount);
end;


//#
procedure HeapCopy(ADest, ASrc: Pointer; ADWordCount: Cardinal);
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

        push esi
        push edi

        mov edi, eax
        mov esi, edx
        rep movsd

        pop edi
        pop esi
end;


procedure HeapCopyByte(ADest, ASrc: Pointer; AByteCount: Cardinal);
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

//      copy dwords as much as possible then complete with the rest of max. 3 bytes

        push esi
        push edi

        mov edi, eax
        mov esi, edx

        mov edx, ecx
        shr ecx, 2      // ecx=ByteCount DIV 4

        rep movsd

        and edx, 3      // edx=ByteCount MOD 4
        jz @end

        mov ecx, edx
        rep movsb

@end:
        pop edi
        pop esi

end;


//#
procedure HeapCopyRev(ADest, ASrc: Pointer; ADWordCount: Cardinal);
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

        push esi
        push edi


        std
        mov edi, eax
        mov esi, edx

        mov eax, ecx
        shl eax, 2 // eax <- block length

        add edi, eax
        sub edi, 4 // edi <- address of end of destination block

        add esi, eax
        sub esi, 4 // esi <- address of end of source block

        rep movsd

        cld

        pop edi
        pop esi
end;


procedure HeapCopyByteRev(ADest, ASrc: Pointer; AByteCount: Cardinal);
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

//      copy dwords as much as possible then complete with the rest of max. 3 bytes

        push esi
        push edi



        std

        mov edi, eax
        mov esi, edx

        mov eax, ecx // eax <- block length

        add edi, eax
        dec edi // edi <- address of end of destination block

        add esi, eax
        dec esi // esi <- address of end of source block

        mov edx, ecx
        shr ecx, 2      // ecx=ByteCount DIV 4

        rep movsd

        and edx, 3      // edx=ByteCount MOD 4
        jz @end

        mov ecx, edx
        rep movsb

@end:

        cld

        pop edi
        pop esi
end;


//#
procedure HeapFill(ADest: Pointer; ADWordPattern: Cardinal; ADWordCount: Cardinal);
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

        push edi

        mov edi, eax
        mov eax, edx

        rep stosd

        pop edi
end;


procedure HeapFillByte(ADest: Pointer; ABytePattern: Byte; AByteCount: Cardinal);
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

//      fill dwords as much as possible then complete with the rest of max. 3 bytes

        push edi

        mov edi, eax
        mov dh, dl
        mov ax, dx
        shl eax, 16
        mov ax, dx      // fill eax with 4 bytes of BytePattern

        mov edx, ecx
        shr ecx, 2      // ecx=ByteCount DIV 4

        rep stosd

        and edx, 3      // edx=ByteCount MOD 4
        jz @end

        mov ecx, edx
        rep stosb

@end:
        pop edi
end;


//#
function HeapCompare(ASrc, ADest: Pointer; ADWordCount: Cardinal): Boolean;
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

//      returns 1 if strings are equal, 0 if not

        push edi
        push esi

        mov edi, eax
        mov esi, edx
        mov @Result, 0

@loop:
        cmpsd
        jnz @exit
        dec ecx
        jnz @loop

        mov @Result, 1
@exit:
        pop esi
        pop edi
end;


function HeapCompareByte(ASrc, ADest: Pointer; AByteCount: Cardinal): Boolean;
asm
//      eax -> first param
//      edx -> second param
//      ecx -> third param

//      compare dwords as much as possible then complete with the rest of max. 3 bytes
//      returns 1 if strings are equal, 0 if not

        push esi
        push edi

        mov edi, eax
        mov esi, edx
        mov @Result, 0

        mov edx, ecx
        shr ecx, 2      // ecx=ByteCount DIV 4
        jz @cmp_rest

@loop1:
        cmpsd
        jnz @exit
        dec ecx
        jnz @loop1

@cmp_rest:
        and edx, 3      // edx=ByteCount MOD 4
        jz @equal

@loop2:
        cmpsb
        jnz @exit
        dec edx
        jnz @loop2

@equal:
        mov @Result, 1

@exit:
        pop edi
        pop esi

end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#######################                                 ######################
//#######################                                 ######################
//##                              INITIALIZATION                              ##
//#######################                                 ######################
//#######################                                 ######################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
procedure Init;
begin
        HeapHandle:= HWindows.HeapCreate(0, 0, 0);

        GetMemoryManager(OldMemoryManager);
        SetMemoryManager(HeapMemoryManager);
end;




//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//#######################                                 ######################
//#######################                                 ######################
//##                                FINALIZATION                              ##
//#######################                                 ######################
//#######################                                 ######################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
//##############################################################################
procedure Final;
begin
        SetMemoryManager(OldMemoryManager);

        HeapDestroy(HeapHandle);
end;

//##############################################################################
//##############################################################################

End.
